// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.
/***/
/*
Calculates the error as the L1-distance between the input mesh and the Lidar mesh data. Output error is presented in a 2D grid representing the X-Y plane
Outputs:
- errorK: error of Gaussian curvature (L1-distance)
- errorH: error of Mean curvature (L1-distance)
- errorCmax: error of Cmax curvature (L1-distance)
- errorCmin: error of Cmin curvature (L1-distance)
- errorSI: error of Shape index (L1-distance)
- errorGL: error of geometric laplacian (L1-distance)
- errorDiAn: error of Dihedral angle (L1-distance)
- validK: flag to indicate Gaussian curvature error whether it is larger or smaller than average + lambda*std
- validH: flag to indicate Mean curvature error whether it is larger or smaller than average + lambda*std
- validCmax: flag to indicate Max curvature error whether it is larger or smaller than average + lambda*std
- validCmin: flag to indicate Min curvature error whether it is larger or smaller than average + lambda*std
- validSI: flag to indicate Shape index error whether it is larger or smaller than average + lambda*std
- validGL: flag to indicate geometric laplacian error whether it is larger or smaller than average + lambda*std
- validDiAn: flag to indicate Dihedral angle error whether it is larger or smaller than average + lambda*std

Inputs:
- indexpoint: buffer containing the vertex to cell grid correspondences
- K1: gaussian curvature for LIDAR data
- H1: mean curvature for LIDAR data
- Cmax1: max curvature for LIDAR data
- Cmin1: min curvature for LIDAR data
- shapeindex1: shape index  for LIDAR data
- GL1: Geometric Laplacian for LIDAR data
- DiAn1: Dihedral angle for input solution.
- K2: gaussian curvature for input solution
- H2: mean curvature for input solution
- Cmax2: max curvature for input solution
- Cmin2: min curvature for input solution
- shapeindex2: shape index  for input solution.
- GL2: Geometric Laplacian for input solution
- DiAn2: Dihedral angle for input solution.
- lambda:specific scale of standard deviation for error truncation according to formula: average + lambda*std
- indexpointcolmajor: buffer containing the vertex to cell grid correspondences sorted by column first indexes (Matlab 1D vectorization)
Version history: 2017-03-28	Marc Bosch	Initial version
*/
/***/
#include <math.h>
#include "mex.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits>


#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{	
    if(nlhs != 14){
        mexErrMsgTxt("Function has 14 return value");
        return;
    }

    if(nrhs !=17){
        mexErrMsgTxt("Usage: [errorK,errorH,errorCmax,errorCmin,errorSI,errorGL,errorDiAn,validK,validH,validCmax,validCmin,validSI,validGL,validDiAn]=mexErrorCurvatureMetricsC(double *indexpoint, double *K1, double *H1, double *Cmax1, double *Cmin1,double *SI1,double *GL1,double DiAn1, double *H2,double *K2, double *Cmax2,double *Cmin2,double *SI2, double *GL2,double DiAn2,double lambda, double indexpointcolmajor)");
        return;
    }
    
    /* get the image buffer*/    
    double *indexpoint = (double*) mxGetPr(prhs[0]);
	double *K1 = (double*) mxGetPr(prhs[1]);
	double *H1 = (double*) mxGetPr(prhs[2]);
	double *Cmax1 = (double*) mxGetPr(prhs[3]);
	double *Cmin1 = (double*) mxGetPr(prhs[4]);
	double *SI1 = (double*) mxGetPr(prhs[5]);
	double *GL1 = (double*) mxGetPr(prhs[6]);
	double *DiAn1 = (double*) mxGetPr(prhs[7]);
    int* dims = (int*) mxGetDimensions(prhs[0]);
    int len = MAX(dims[0],dims[1]);
	double *K2 = (double*) mxGetPr(prhs[8]);
	double *H2 = (double*) mxGetPr(prhs[9]);
	double *Cmax2 = (double*) mxGetPr(prhs[10]);
	double *Cmin2 = (double*) mxGetPr(prhs[11]);
	double *SI2 = (double*) mxGetPr(prhs[12]);
	double *GL2 = (double*) mxGetPr(prhs[13]);
	double *DiAn2 = (double*) mxGetPr(prhs[14]);
	int* dimsgl = (int*) mxGetDimensions(prhs[13]);
	int len1 = MAX(dims[0],dims[1]);
	int len2 = MIN(dims[0],dims[1]);
    double mean_K=0,mean_H=0,mean_Cmax=0,mean_Cmin=0,mean_SI=0,mean_GL=0, mean_DiAn=0;
	double std_K=0,std_H=0,std_Cmax=0,std_Cmin=0,std_SI=0,std_GL=0,std_DiAn=0;
	double* plambda = mxGetPr(prhs[15]);
 	double lambda = *plambda;
	double *indexpointcolmajor = (double*) mxGetPr(prhs[16]);
	//allocate outputs
	plhs[0] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorK = mxGetPr(plhs[0]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorK[i] = 0;//dims[0];  
	plhs[1] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorH = mxGetPr(plhs[1]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorH[i] = 0;//dims[0]; 
	plhs[2] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorCmax = mxGetPr(plhs[2]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorCmax[i] = 0;//dims[0]; 
	plhs[3] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorCmin = mxGetPr(plhs[3]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorCmin[i] = 0;//dims[0]; 
	plhs[4] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorSI = mxGetPr(plhs[4]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorSI[i] = 0;//dims[0]; 
	plhs[5] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorGL = mxGetPr(plhs[5]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorGL[i] = 0;//dims[0]; 
	plhs[6] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *errorDiAn = mxGetPr(plhs[6]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        errorDiAn[i] = 0;//dims[0]; 
	plhs[7] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *validK = mxGetPr(plhs[7]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validK[i] = 0;//dims[0]; 
	plhs[8] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *validH = mxGetPr(plhs[8]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validH[i] = 0;//dims[0]; 
	plhs[9] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *validCmax = mxGetPr(plhs[9]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validCmax[i] = 0;//dims[0]; 
	plhs[10] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *validCmin = mxGetPr(plhs[10]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validCmin[i] = 0;//dims[0]; 
	plhs[11] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *validSI = mxGetPr(plhs[11]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validSI[i] = 0;//dims[0]; 
	plhs[12] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
    double *validGL = mxGetPr(plhs[12]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validGL[i] = 0;//dims[0]; 
	plhs[13] = mxCreateDoubleMatrix(dims[0],dims[1], mxREAL);
	double *validDiAn = mxGetPr(plhs[13]);  
    for (int i=0; i < dims[0]*dims[1]; i++)
        validDiAn[i] = 0;//dims[0];
	int lenDiAn=0;
	for (int i=0;i<len;i++)
	{
		errorK[i] = abs(K1[(int)indexpoint[i]-1]-K2[i]);
		errorH[i] = abs(H1[(int)indexpoint[i]-1]-H2[i]);
		errorCmax[i] = abs(Cmax1[(int)indexpoint[i]-1]-Cmax2[i]);
		errorCmin[i] = abs(Cmin1[(int)indexpoint[i]-1]-Cmin2[i]);
		errorSI[i] = abs(SI1[(int)indexpoint[i]-1]-SI2[i]);
		errorGL[i] = ((GL1[(int)(indexpointcolmajor[i]-1)*3+0]-GL2[i*3+0])*
			(GL1[(int)(indexpointcolmajor[i]-1)*3+0]-GL2[i*3+0])+
			(GL1[(int)(indexpointcolmajor[i]-1)*3+1]-GL2[i*3+1])*
			(GL1[(int)(indexpointcolmajor[i]-1)*3+1]-GL2[i*3+1])+
			(GL1[(int)(indexpointcolmajor[i]-1)*3+2]-GL2[i*3+2])*
			(GL1[(int)(indexpointcolmajor[i]-1)*3+2]-GL2[i*3+2]));
		
		validDiAn[i]=1;
		if(DiAn1[(int)indexpointcolmajor[i]]!=-1000 && DiAn2[i]!=-1000)
		{
			errorDiAn[i] = abs(DiAn1[(int)indexpointcolmajor[i]]-DiAn2[i]);
			mean_DiAn+=errorDiAn[i];
			std_DiAn+=(errorDiAn[i]*errorDiAn[i]);
			lenDiAn++;
		}
		else
		{
			errorDiAn[i]=180;
			validDiAn[i]=0;
		}

		mean_K+=errorK[i];
		std_K+=(errorK[i]*errorK[i]);
		mean_H+=errorH[i];
		std_H+=(errorH[i]*errorH[i]);
		mean_Cmax+=errorCmax[i];
		std_Cmax+=(errorCmax[i]*errorCmax[i]);
		mean_Cmin+=errorCmin[i];
		std_Cmin+=(errorCmin[i]*errorCmin[i]);
		mean_SI+=errorSI[i];
		std_SI+=(errorSI[i]*errorSI[i]);
		mean_GL+=errorGL[i];
		std_GL+=(errorGL[i]*errorGL[i]);
		
	}
	len = (len<2) ? 2 : len;
	double temp=mean_K;
    mean_K = mean_K/len;
    std_K =  sqrt((std_K-temp*mean_K)/(len-1));
	temp=mean_H;
    mean_H = mean_H/len;
    std_H =  sqrt((std_H-temp*mean_H)/(len-1));
	temp=mean_Cmax;
    mean_Cmax = mean_Cmax/len;
    std_Cmax =  sqrt((std_Cmax-temp*mean_Cmax)/(len-1));
	temp=mean_Cmin;
    mean_Cmin = mean_Cmin/len;
    std_Cmin =  sqrt((std_Cmin-temp*mean_Cmin)/(len-1));
	temp=mean_SI;
    mean_SI = mean_SI/len;
    std_SI =  sqrt((std_SI-temp*mean_SI)/(len-1));
	temp=mean_GL;
    mean_GL = mean_GL/len;
    std_GL =  sqrt((std_GL-temp*mean_GL)/(len-1));
	temp=mean_DiAn;
    mean_DiAn = mean_DiAn/lenDiAn;
    std_DiAn =  sqrt((std_DiAn-temp*mean_DiAn)/(lenDiAn-1));

	for (int i=0;i<len;i++)
	{
		validK[i]=1;
		if (errorK[i]>mean_K+lambda*std_K)
			validK[i]=0;
		validH[i]=1;
		if (errorH[i]>mean_H+lambda*std_H)
			validH[i]=0;
		validCmax[i]=1;
		if (errorCmax[i]>mean_Cmax+lambda*std_Cmax)
			validCmax[i]=0;
		validCmin[i]=1;
		if (errorCmin[i]>mean_Cmin+lambda*std_Cmin)
			validCmin[i]=0;
		validSI[i]=1;
		if (errorSI[i]>mean_SI+lambda*std_SI)
			validSI[i]=0;
		validGL[i]=1;
		if (errorGL[i]>mean_GL+lambda*std_GL)
			validGL[i]=0;
//		validDiAn[i]=1;
		if (errorDiAn[i]>mean_DiAn+lambda*std_DiAn)
			validDiAn[i]=0;
	}   
   
   
}
